import { codegen } from '@graphql-codegen/core';
import * as typescriptPlugin from '@graphql-codegen/typescript';
import * as typescriptOperationsPlugin from '@graphql-codegen/typescript-operations';
import { CodeFileLoader } from '@graphql-toolkit/code-file-loader';
import { loadDocuments, loadSchema } from '@graphql-toolkit/core';
import { JsonFileLoader } from '@graphql-toolkit/json-file-loader';
import chokidar from 'chokidar';
import crypto from 'crypto';
import fs from 'fs';
import { GatsbyNode } from 'gatsby';
import { getIntrospectionQuery, graphql, parse, printSchema } from 'graphql';
import debounce from 'lodash.debounce';
import path from 'path';

import { fixStaticQuery, fixUseStaticQuery } from './fixers';
import { PluginOptions } from './types';

const resolve = (...paths: string[]) => path.resolve(process.cwd(), ...paths);
const log = (message: string) =>
  // tslint:disable-next-line: no-console
  console.log(`[gatsby-plugin-typegen] ${message}`);

const DOCUMENTS_GLOB = resolve('src/**/*.{ts,tsx}');
const DEFAULT_SCHEMA_OUTPUT_PATH = resolve(
  '.cache/caches/gatsby-plugin-typegen/schema.json',
);
const DEFAULT_TYPE_DEFS_OUTPUT_PATH = resolve(
  'src/__generated__/types/graphql.ts',
);

export const onPostBootstrap: GatsbyNode['onPostBootstrap'] = async (
  { store },
  options,
) => {
  const {
    schemaOutputPath = DEFAULT_SCHEMA_OUTPUT_PATH,
    typeDefsOutputPath = DEFAULT_TYPE_DEFS_OUTPUT_PATH,
    autoFix = true,
  } = options as PluginOptions;

  let cache = '';

  const extractSchema = async () => {
    const { schema } = store.getState();
    const introspectionQuery = getIntrospectionQuery({ descriptions: true });
    const { data: fullSchema } = await graphql(schema, introspectionQuery);
    const output = JSON.stringify(fullSchema, null, 2);
    const sha1sum = crypto
      .createHash('sha1')
      .update(output)
      .digest('hex');

    if (cache !== sha1sum) {
      cache = sha1sum;

      await fs.promises.writeFile(schemaOutputPath, output, 'utf-8');
      log(`Schema file extracted to ${schemaOutputPath}!`);
    }
  };

  // Wait for first extraction
  await extractSchema();
  const schemaAst = await loadSchema(schemaOutputPath, {
    loaders: [new JsonFileLoader()],
  });
  const documents = await loadDocuments(resolve(DOCUMENTS_GLOB), {
    loaders: [new CodeFileLoader()],
  });
  log('Documents are loaded');

  const plugins: Array<{ typescript: [] } | { typescriptOperations: [] }> = [
    { typescript: [] },
    { typescriptOperations: [] },
  ];

  const config = {
    schema: parse(printSchema(schemaAst)),
    schemaAst,
    filename: typeDefsOutputPath,
    plugins,
    pluginMap: {
      typescript: typescriptPlugin,
      typescriptOperations: typescriptOperationsPlugin,
    },
    config: {
      avoidOptionals: true,
      maybeValue: 'T',
      namingConvention: {
        enumValues: 'keep',
        transformUnderscore: false,
      },
    },
  };

  const writeTypeDefinition = debounce(async () => {
    const output = await codegen({
      ...config,
      documents,
    });

    await fs.promises.mkdir(path.dirname(typeDefsOutputPath), {
      recursive: true,
    });

    await fs.promises.writeFile(typeDefsOutputPath, output, 'utf-8');
    log(`Type definitions are generated into ${typeDefsOutputPath}`);
  }, 1000);

  const fixDocuments = async (filePath: string) => {
    const code = await fs.promises.readFile(filePath, 'utf-8');
    const fixed = fixUseStaticQuery(fixStaticQuery(code));

    if (code !== fixed) {
      fs.promises.writeFile(filePath, fixed, 'utf-8');
    }
  };

  const onWatch = async (filePath: string) => {
    if (filePath === typeDefsOutputPath) {
      return;
    }

    const changed = documents.findIndex(
      document => document.location === filePath,
    );

    if (changed !== -1) {
      documents[changed] = (
        await loadDocuments(filePath, {
          loaders: [new CodeFileLoader()],
        })
      )[0];
    }

    writeTypeDefinition();

    if (autoFix && filePath !== schemaOutputPath) {
      fixDocuments(filePath);
    }
  };

  const watcher = chokidar.watch([schemaOutputPath, DOCUMENTS_GLOB]);

  watcher.on('add', onWatch).on('change', onWatch);

  store.subscribe(extractSchema);
};
